# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

import collections.abc as cabc

import pytest

import capellambse


def test_state_realized_states(model: capellambse.MelodyModel) -> None:
    state = model.by_uuid("43932114-8ad4-4074-b2a9-b0d55b8d027b")
    real = model.by_uuid("461de288-fa3f-45bd-9975-8ebe1aec85d9")
    real1 = model.by_uuid("757c825f-f522-4d90-b3af-071a4cec3f06")

    assert state.realized_states[0].realizing_states[0] == state
    assert state.realized_states == [real, real1]


def test_state_realizing_states(model: capellambse.MelodyModel) -> None:
    state = model.by_uuid("461de288-fa3f-45bd-9975-8ebe1aec85d9")
    real = model.by_uuid("43932114-8ad4-4074-b2a9-b0d55b8d027b")

    assert state.realizing_states == [real]


@pytest.mark.parametrize(
    "attr,uuids",
    [
        ("entries", ("0e0164c3-076e-42c1-8f82-7a43ab84385c",)),
        ("do_activity", ("8bcb11e6-443b-4b92-bec2-ff1d87a224e7",)),
        (
            "exits",
            (
                "ab016025-d456-4c2d-8359-d50cf9de3825",
                "f33bd79c-fa90-4c09-942b-581dc7f07d84",
                "04babdbf-6cf6-4846-a207-bf27cfc8eb32",
            ),
        ),
    ],
)
def test_state_attributes(
    model_5_2: capellambse.MelodyModel, attr: str, uuids: cabc.Iterable[str]
) -> None:
    state = model_5_2.by_uuid("6c48b9c5-0d43-4a43-9e9d-9559cb52c83e")
    expected_functions = [model_5_2.by_uuid(uuid) for uuid in uuids]

    functions = getattr(state, attr)

    assert functions == expected_functions



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Tools for the Physical Architecture layer.

.. diagram:: [CDB] Physical Architecture [Ontology]
"""
from __future__ import annotations

from .. import common as c
from .. import crosslayer, diagram, modeltypes
from ..crosslayer import capellacommon, cs, fa
from . import la

XT_ARCH = "org.polarsys.capella.core.data.pa:PhysicalArchitecture"

XT_LA_COMP_REAL = (
    "org.polarsys.capella.core.data.pa:LogicalComponentRealization"
)


@c.xtype_handler(XT_ARCH)
class PhysicalFunction(fa.Function):
    """A physical function on the Physical Architecture layer."""

    _xmltag = "ownedPhysicalFunctions"

    owner: c.Accessor[PhysicalComponent]
    realized_logical_functions = c.TypecastAccessor(
        la.LogicalFunction, "realized_functions"
    )


@c.xtype_handler(XT_ARCH)
class PhysicalFunctionPkg(c.GenericElement):
    """A logical component package."""

    _xmltag = "ownedFunctionPkg"

    functions = c.DirectProxyAccessor(PhysicalFunction, aslist=c.ElementList)

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class PhysicalComponent(cs.Component):
    """A physical component on the Physical Architecture layer."""

    _xmltag = "ownedPhysicalComponents"

    nature = c.EnumAttributeProperty("nature", modeltypes.Nature)
    kind = c.EnumAttributeProperty(
        "kind", modeltypes.Kind, default=modeltypes.Kind.UNSET
    )

    allocated_functions = c.LinkAccessor[PhysicalFunction](
        "ownedFunctionalAllocation",
        fa.XT_FCALLOC,
        aslist=c.ElementList,
        attr="targetElement",
    )
    realized_logical_components = c.TypecastAccessor(
        la.LogicalComponent,
        "realized_components",
    )

    owned_components: c.Accessor
    deploying_components: c.Accessor

    @property
    def deployed_components(
        self,
    ) -> c.ElementList[PhysicalComponent]:
        items = [
            cmp.type._element
            for part in self.parts
            for cmp in part.deployed_parts
        ]
        return c.ElementList(self._model, items, PhysicalComponent)

    @property
    def components(self) -> c.ElementList[PhysicalComponent]:
        return self.deployed_components + self.owned_components

    functions = c.DeprecatedAccessor[PhysicalFunction]("allocated_functions")


@c.xtype_handler(XT_ARCH)
class PhysicalComponentPkg(c.GenericElement):
    """A logical component package."""

    _xmltag = "ownedPhysicalComponentPkg"

    components = c.DirectProxyAccessor(PhysicalComponent, aslist=c.ElementList)
    exchanges = c.DirectProxyAccessor(
        fa.ComponentExchange, aslist=c.ElementList
    )
    state_machines = c.DirectProxyAccessor(
        capellacommon.StateMachine, aslist=c.ElementList
    )

    packages: c.Accessor


@c.xtype_handler(None)
class PhysicalArchitecture(crosslayer.BaseArchitectureLayer):
    """Provides access to the Physical Architecture layer of the model."""

    root_component = c.AttributeMatcherAccessor(
        PhysicalComponent,
        attributes={"is_actor": False},
        rootelem=PhysicalComponentPkg,
    )
    root_function = c.DirectProxyAccessor(
        PhysicalFunction, rootelem=PhysicalFunctionPkg
    )

    function_package = c.DirectProxyAccessor(PhysicalFunctionPkg)
    component_package = c.DirectProxyAccessor(PhysicalComponentPkg)
    capability_package = c.DirectProxyAccessor(la.CapabilityRealizationPkg)

    all_functions = c.DeepProxyAccessor(
        PhysicalFunction,
        aslist=c.ElementList,
        rootelem=PhysicalFunctionPkg,
    )
    all_capabilities = c.DeepProxyAccessor(
        la.CapabilityRealization, aslist=c.ElementList
    )
    all_components = c.DeepProxyAccessor(
        PhysicalComponent, aslist=c.ElementList
    )
    all_actors = property(
        lambda self: self._model.search(PhysicalComponent).by_is_actor(True)
    )

    all_function_exchanges = c.DeepProxyAccessor(
        fa.FunctionalExchange,
        aslist=c.ElementList,
        rootelem=[PhysicalFunctionPkg, PhysicalFunction],
    )
    all_physical_paths = c.DeepProxyAccessor(
        cs.PhysicalPath,
        aslist=c.ElementList,
        rootelem=PhysicalComponentPkg,
    )
    all_component_exchanges = c.DeepProxyAccessor(
        fa.ComponentExchange,
        aslist=c.ElementList,
        rootelem=PhysicalComponentPkg,
    )

    all_physical_exchanges = c.DeepProxyAccessor(
        fa.FunctionalExchange,
        aslist=c.ElementList,
        rootelem=[PhysicalFunctionPkg, PhysicalFunction],
    )
    all_physical_links = c.DeepProxyAccessor(
        cs.PhysicalLink, aslist=c.ElementList
    )
    all_functional_chains = property(
        lambda self: self._model.search(fa.FunctionalChain, below=self)
    )

    diagrams = diagram.DiagramAccessor(
        "Physical Architecture", cacheattr="_MelodyModel__diagram_cache"
    )


c.set_accessor(
    la.LogicalComponent,
    "realizing_physical_components",
    c.ReferenceSearchingAccessor(
        PhysicalComponent, "realized_logical_components", aslist=c.ElementList
    ),
)
c.set_accessor(
    la.LogicalFunction,
    "realizing_physical_functions",
    c.ReferenceSearchingAccessor(
        PhysicalFunction, "realized_logical_functions", aslist=c.ElementList
    ),
)
c.set_accessor(
    PhysicalComponent,
    "deploying_components",
    c.ReferenceSearchingAccessor(
        PhysicalComponent, "deployed_components", aslist=c.ElementList
    ),
)
c.set_accessor(
    PhysicalFunction,
    "owner",
    c.ReferenceSearchingAccessor(PhysicalComponent, "allocated_functions"),
)
c.set_accessor(
    PhysicalFunction,
    "packages",
    c.DirectProxyAccessor(PhysicalFunctionPkg, aslist=c.ElementList),
)
c.set_self_references(
    (PhysicalComponent, "owned_components"),
    (PhysicalComponentPkg, "packages"),
    (PhysicalFunction, "functions"),
    (PhysicalFunctionPkg, "packages"),
)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import json
import pathlib

import pytest
from lxml import etree

import capellambse
import capellambse.diagram
from capellambse.svg import (
    SVGDiagram,
    decorations,
    generate,
    helpers,
    style,
    symbols,
)

TEST_LAB = "[LAB] Wizzard Education"
TEST_DIAGS = [
    TEST_LAB,
    "[OAB] Operational Context",
    "[OCB] Operational Capabilities",
    "[OPD] Obtain food via hunting",
    "[MSM] States of Functional Human Being",
    "[SAB] System",
    "[OEBD] Operational Context",
    "[PAB] Physical System",
    "[LDFB] Test flow",
    "[CC] Capability",
]
TEST_DECO = set(style.STATIC_DECORATIONS.keys()) - {"__GLOBAL__"}
FREE_SYMBOLS = {
    "OperationalCapabilitySymbol",
    "AndControlNodeSymbol",
    "ItControlNodeSymbol",
    "OrControlNodeSymbol",
    "FinalStateSymbol",
    "InitialPseudoStateSymbol",
    "TerminatePseudoStateSymbol",
    "StickFigureSymbol",
}


@pytest.fixture(name="tmp_json")
def tmp_json_fixture(
    model: capellambse.MelodyModel, tmp_path: pathlib.Path
) -> pathlib.Path:
    """Return tmp path of diagram json file."""
    dest = tmp_path / (TEST_LAB + ".json")
    diagram = model.diagrams.by_name(TEST_LAB).render(None)
    diagram_json = capellambse.diagram.DiagramJSONEncoder().encode(diagram)
    dest.write_text(diagram_json)
    return dest


class TestSVG:
    def test_diagram_meta_data_attributes(
        self, tmp_json: pathlib.Path
    ) -> None:
        diag_meta = generate.DiagramMetadata.from_dict(
            json.loads(tmp_json.read_text())
        )
        assert diag_meta.name == TEST_LAB
        assert diag_meta.pos == (15, 15)
        assert diag_meta.size == (1162, 611)
        assert diag_meta.viewbox == "15 15 1162 611"
        assert diag_meta.class_ == "Logical Architecture Blank"

    def test_diagram_from_json_path_componentports(
        self, tmp_json: pathlib.Path
    ) -> None:
        tree = etree.fromstring(
            SVGDiagram.from_json_path(tmp_json).to_string()
        )

        cp_in_exists: bool = False
        cp_inout_exists: bool = False
        cp_out_exists: bool = False
        cp_unset_exists: bool = False
        cp_reference_exists: bool = False

        for item in tree.iter():
            # The class CP should not exist anymore as it has been replaced
            # with CP_IN, CP_OUT, CP_UNSET or CP_INOUT
            assert item.get("class") != "Box CP"

            # Check that the classes CP_IN, CP_OUT, CP_UNSET and CP_INOUT exist
            if item.get("class") == "Box CP_IN":
                cp_in_exists = True
            elif item.get("class") == "Box CP_OUT":
                cp_out_exists = True
            elif item.get("class") == "Box CP_INOUT":
                cp_inout_exists = True
            elif item.get("class") == "Box CP_UNSET":
                cp_unset_exists = True

            # Check that reference symbol for CP exists
            if (
                item.tag == "{http://www.w3.org/2000/svg}symbol"
                and item.get("id") == "ComponentPortSymbol"
            ):
                cp_reference_exists = True

        assert cp_in_exists
        assert cp_out_exists
        assert cp_inout_exists
        assert cp_unset_exists
        assert cp_reference_exists

    @pytest.fixture
    def tmp_svg(self, tmp_path: pathlib.Path) -> SVGDiagram:
        name = "Test svg"
        meta = generate.DiagramMetadata(
            pos=(0, 0), size=(1, 1), name=name, class_="TEST"
        )
        svg = SVGDiagram(meta, [])
        svg.drawing.filename = str(tmp_path / name)
        return svg

    def test_diagram_saves(self, tmp_svg: SVGDiagram) -> None:
        tmp_svg.save_drawing()
        assert pathlib.Path(tmp_svg.drawing.filename).is_file()

    def test_base_css_styles(self, tmp_json: pathlib.Path) -> None:
        tree = etree.fromstring(
            SVGDiagram.from_json_path(tmp_json).to_string()
        )
        style_ = tree.get("style")
        assert style_

    @pytest.mark.parametrize("diagram_name", TEST_DIAGS)
    def test_diagram_decorations(
        self, model: capellambse.MelodyModel, diagram_name: str
    ):
        """Test diagrams get rendered successfully."""
        diag = model.diagrams.by_name(diagram_name)
        diag.render("svg")


class TestDecoFactory:
    @pytest.mark.parametrize(
        "class_",
        ["LogicalComponentSymbol", "LogicalHumanActorSymbol", "EntitySymbol"],
    )
    def test_deco_factory_contains_styling_for_given_styleclass(
        self, class_: str
    ):
        assert class_ in decorations.deco_factories

    def test_deco_factory_returns_symbol_factory_for_given_styleclass(self):
        assert decorations.deco_factories["PortSymbol"] is symbols.port_symbol

    @pytest.mark.parametrize(
        "class_", ["ImaginaryClassSymbol", "NothingSymbol"]
    )
    def test_deco_factory_logs_error_when_not_containing_given_styleclass(
        self, caplog, class_: str
    ):
        assert class_ not in decorations.deco_factories
        assert (
            decorations.deco_factories[class_]
            is decorations.deco_factories["ErrorSymbol"]
        )
        with caplog.at_level(0, logger="decorations"):
            assert (
                caplog.messages[-1] == f"{class_} wasn't found in factories."
            )

    @pytest.mark.parametrize("attr", ["start", "translate", "offsets"])
    def test_making_linear_gradient_faulty_cases_raise_ValueError(
        self, attr: str
    ):
        params = {
            "id_": "test",
            attr: (0, 0, 0),
        }
        with pytest.raises(ValueError):
            symbols._make_lgradient(**params)

    def test_making_linear_gradient_with_translate(self):
        gradient = symbols._make_lgradient("test", translate=(0, 0))
        assert gradient.attribs.get("gradientTransform") == "translate(0 0)"


class TestSVGHelpers:
    def test_check_for_horizontal_overflow_recognizes_tabs_and_breaks(
        self,
    ) -> None:
        lines, margin, max_text_width = helpers.check_for_horizontal_overflow(
            "             • item 1\n             • item 2", 100, 0, 0
        )
        assert lines == ["             • item 1", "             • item 2"]
        assert 10 <= margin < 13
        for line in lines:
            assert capellambse.helpers.extent_func(line)[0] <= max_text_width



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0


from .. import common as c
from . import capellacore

XT_CAP2PROC = (
    "org.polarsys.capella.core.data.interaction"
    ":FunctionalChainAbstractCapabilityInvolvement"
)
XT_CAP_REAL = (
    "org.polarsys.capella.core.data.interaction:AbstractCapabilityRealization"
)


@c.xtype_handler(None)
class Execution(c.GenericElement):
    """An execution."""

    start = c.AttrProxyAccessor(c.GenericElement, "start")
    finish = c.AttrProxyAccessor(c.GenericElement, "finish")


@c.xtype_handler(None)
class StateFragment(Execution):
    """A state fragment."""

    function = c.AttrProxyAccessor(c.GenericElement, "relatedAbstractFunction")


@c.xtype_handler(None)
class CombinedFragment(Execution):
    """A combined fragment."""

    operator = c.AttributeProperty("operator", optional=True)
    operands = c.AttrProxyAccessor(
        c.GenericElement, "referencedOperands", aslist=c.ElementList
    )


@c.xtype_handler(None)
class InstanceRole(c.GenericElement):
    """An instance role."""

    instance = c.AttrProxyAccessor[c.GenericElement](
        None, "representedInstance"
    )


@c.xtype_handler(None)
class SequenceMessage(c.GenericElement):
    """A sequence message."""

    source = c.AttrProxyAccessor(c.GenericElement, "sendingEnd")
    target = c.AttrProxyAccessor(c.GenericElement, "receivingEnd")


class Event(c.GenericElement):
    """Abstract super class of all events in a Scenario."""


class EventOperation(Event):
    """Abstract super class for events about operations."""

    operation = c.AttrProxyAccessor(c.GenericElement, "operation")


@c.xtype_handler(None)
class ExecutionEvent(Event):
    """An execution event."""


@c.xtype_handler(None)
class EventSentOperation(EventOperation):
    """An event-sent operation."""


@c.xtype_handler(None)
class EventReceiptOperation(EventOperation):
    """An event-receipt operation."""


@c.xtype_handler(None)
class Scenario(c.GenericElement):
    """A scenario that holds instance roles."""

    instance_roles = c.DirectProxyAccessor[c.GenericElement](
        InstanceRole, aslist=c.ElementList
    )
    messages = c.DirectProxyAccessor(SequenceMessage, aslist=c.ElementList)
    events = c.RoleTagAccessor("ownedEvents", aslist=c.MixedElementList)
    fragments = c.RoleTagAccessor(
        "ownedInteractionFragments", aslist=c.MixedElementList
    )
    time_lapses = c.RoleTagAccessor(
        "ownedTimeLapses", aslist=c.MixedElementList
    )
    postcondition = c.AttrProxyAccessor(
        capellacore.Constraint, "postCondition"
    )
    precondition = c.AttrProxyAccessor(capellacore.Constraint, "preCondition")


class InteractionFragment(c.GenericElement):
    """Abstract super class of all interaction fragments in a Scenario."""

    covered = c.AttrProxyAccessor[c.GenericElement](
        None, "coveredInstanceRoles", aslist=c.MixedElementList
    )


@c.xtype_handler(None)
class ExecutionEnd(InteractionFragment):
    """An end for an execution."""

    event = c.AttrProxyAccessor[Event](None, "event")


@c.xtype_handler(None)
class FragmentEnd(InteractionFragment):
    """An end for a fragment."""


@c.xtype_handler(None)
class InteractionOperand(InteractionFragment):
    """An interaction-operand."""

    guard = c.AttrProxyAccessor(capellacore.Constraint, "guard")


@c.xtype_handler(None)
class InteractionState(InteractionFragment):
    """An interaction-state."""

    state = c.AttrProxyAccessor(c.GenericElement, "relatedAbstractState")
    function = c.AttrProxyAccessor(c.GenericElement, "relatedAbstractFunction")


@c.xtype_handler(None)
class MessageEnd(InteractionFragment):
    """A message-end."""

    event = c.AttrProxyAccessor[Event](None, "event")


class Exchange(c.GenericElement):
    """An abstract Exchange."""

    source = c.ParentAccessor(c.GenericElement)


@c.xtype_handler(None)
class AbstractCapabilityExtend(Exchange):
    """An AbstractCapabilityExtend."""

    _xmltag = "extends"

    source = c.ParentAccessor(c.GenericElement)
    target = c.AttrProxyAccessor(c.GenericElement, "extended")


@c.xtype_handler(None)
class AbstractCapabilityInclude(Exchange):
    """An AbstractCapabilityInclude."""

    _xmltag = "includes"

    source = c.ParentAccessor(c.GenericElement)
    target = c.AttrProxyAccessor(c.GenericElement, "included")


@c.xtype_handler(None)
class AbstractCapabilityGeneralization(Exchange):
    """An AbstractCapabilityGeneralization."""

    _xmltag = "superGeneralizations"

    source = c.ParentAccessor(c.GenericElement)
    target = c.AttrProxyAccessor(c.GenericElement, "super")


class AbstractInvolvement(c.GenericElement):
    """An abstract Involvement."""

    source = c.ParentAccessor(c.GenericElement)
    target = c.AttrProxyAccessor(c.GenericElement, "involved")

    involved = c.AttrProxyAccessor(c.GenericElement, "involved")

    @property
    def name(self) -> str:  # type: ignore[override]
        """Return the name."""
        direction = ""
        if self.involved is not None:
            direction = f" to {self.involved.name} ({self.involved.uuid})"

        return f"[{self.__class__.__name__}]{direction}"


@c.xtype_handler(None)
class AbstractFunctionAbstractCapabilityInvolvement(AbstractInvolvement):
    """An abstract CapabilityInvolvement linking to SystemFunctions."""



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Provides easy access to the Polarsys Capella PVMT extension.

The public API of this submodule uses raw LXML elements. For a more
object oriented and user friendly way to access property values in a
model, see the :class:`capellambse.MelodyModel` class.
"""

from .exceptions import ScopeError
from .model import PVMTExtension, load_pvmt_from_model



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Tools for the Logical Architecture layer.

.. diagram:: [CDB] LA ORM
"""
from __future__ import annotations

from .. import common as c
from .. import crosslayer, diagram
from ..crosslayer import capellacommon, capellacore, cs, fa, interaction
from . import ctx

XT_ARCH = "org.polarsys.capella.core.data.la:LogicalArchitecture"


@c.xtype_handler(XT_ARCH)
class LogicalFunction(fa.Function):
    """A logical function on the Logical Architecture layer."""

    _xmltag = "ownedLogicalFunctions"

    realized_system_functions = c.TypecastAccessor(
        ctx.SystemFunction, "realized_functions"
    )
    owner: c.Accessor[LogicalComponent]


@c.xtype_handler(XT_ARCH)
class LogicalFunctionPkg(c.GenericElement):
    """A logical function package."""

    _xmltag = "ownedFunctionPkg"

    functions = c.DirectProxyAccessor(LogicalFunction, aslist=c.ElementList)

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class LogicalComponent(cs.Component):
    """A logical component on the Logical Architecture layer."""

    _xmltag = "ownedLogicalComponents"

    allocated_functions = c.LinkAccessor[LogicalFunction](
        "ownedFunctionalAllocation",
        fa.XT_FCALLOC,
        aslist=c.ElementList,
        attr="targetElement",
    )
    realized_system_components = c.TypecastAccessor(
        ctx.SystemComponent,
        "realized_components",
    )

    components: c.Accessor
    functions = c.DeprecatedAccessor[LogicalFunction]("allocated_functions")


@c.xtype_handler(XT_ARCH)
class LogicalComponentPkg(c.GenericElement):
    """A logical component package."""

    _xmltag = "ownedLogicalComponentPkg"

    components = c.DirectProxyAccessor(LogicalComponent, aslist=c.ElementList)
    state_machines = c.DirectProxyAccessor(
        capellacommon.StateMachine, aslist=c.ElementList
    )
    exchanges = c.DirectProxyAccessor(
        fa.ComponentExchange, aslist=c.ElementList
    )

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class CapabilityRealization(c.GenericElement):
    """A capability."""

    _xmltag = "ownedCapabilityRealizations"

    owned_chains = c.DirectProxyAccessor(
        fa.FunctionalChain, aslist=c.ElementList
    )
    involved_functions = c.LinkAccessor[LogicalFunction](
        "ownedAbstractFunctionAbstractCapabilityInvolvements",
        interaction.AbstractFunctionAbstractCapabilityInvolvement,
        aslist=c.ElementList,
        attr="involved",
    )
    involved_chains = c.LinkAccessor[fa.FunctionalChain](
        "ownedFunctionalChainAbstractCapabilityInvolvements",
        interaction.XT_CAP2PROC,
        aslist=c.ElementList,
        attr="involved",
    )
    involved_components = c.LinkAccessor[LogicalComponent](
        "ownedCapabilityInvolvements",
        ctx.CapabilityInvolvement,
        aslist=c.MixedElementList,
        attr="involved",
    )
    realized_capabilities = c.LinkAccessor[ctx.Capability](
        None,  # FIXME fill in tag
        interaction.XT_CAP_REAL,
        aslist=c.ElementList,
        attr="targetElement",
    )

    postcondition = c.AttrProxyAccessor(
        capellacore.Constraint, "postCondition"
    )
    precondition = c.AttrProxyAccessor(capellacore.Constraint, "preCondition")
    scenarios = c.DirectProxyAccessor(
        interaction.Scenario, aslist=c.ElementList
    )
    states = c.AttrProxyAccessor(
        capellacommon.State, "availableInStates", aslist=c.ElementList
    )

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class CapabilityRealizationPkg(c.GenericElement):
    """A capability package that can hold capabilities."""

    _xmltag = "ownedAbstractCapabilityPkg"

    capabilities = c.DirectProxyAccessor(
        CapabilityRealization, aslist=c.ElementList
    )

    packages: c.Accessor


@c.xtype_handler(None)
class LogicalArchitecture(crosslayer.BaseArchitectureLayer):
    """Provides access to the LogicalArchitecture layer of the model."""

    root_component = c.AttributeMatcherAccessor(
        LogicalComponent,
        attributes={"is_actor": False},
        rootelem=LogicalComponentPkg,
    )
    root_function = c.DirectProxyAccessor(
        LogicalFunction, rootelem=LogicalFunctionPkg
    )

    function_package = c.DirectProxyAccessor(LogicalFunctionPkg)
    component_package = c.DirectProxyAccessor(LogicalComponentPkg)
    capability_package = c.DirectProxyAccessor(CapabilityRealizationPkg)

    all_functions = c.DeepProxyAccessor(
        LogicalFunction,
        aslist=c.ElementList,
        rootelem=LogicalFunctionPkg,
    )
    all_capabilities = c.DeepProxyAccessor(
        CapabilityRealization, aslist=c.ElementList
    )
    all_components = (  # maybe this should exclude .is_actor
        c.DeepProxyAccessor(LogicalComponent, aslist=c.ElementList)
    )
    all_actors = property(
        lambda self: self._model.search(LogicalComponent).by_is_actor(True)
    )
    all_functional_chains = property(
        lambda self: self._model.search(fa.FunctionalChain, below=self)
    )

    actor_exchanges = c.DirectProxyAccessor(
        fa.ComponentExchange,
        aslist=c.ElementList,
        rootelem=LogicalComponentPkg,
    )
    component_exchanges = c.DeepProxyAccessor(
        fa.ComponentExchange,
        aslist=c.ElementList,
        rootelem=[LogicalComponentPkg, LogicalComponent],
    )

    all_function_exchanges = c.DeepProxyAccessor(
        fa.FunctionalExchange,
        aslist=c.ElementList,
        rootelem=[LogicalFunctionPkg, LogicalFunction],
    )
    all_component_exchanges = c.DeepProxyAccessor(
        fa.ComponentExchange, aslist=c.ElementList
    )

    diagrams = diagram.DiagramAccessor(
        "Logical Architecture", cacheattr="_MelodyModel__diagram_cache"
    )


c.set_accessor(
    ctx.Capability,
    "realizing_capabilities",
    c.ReferenceSearchingAccessor(
        CapabilityRealization, "realized_capabilities", aslist=c.ElementList
    ),
)
c.set_accessor(
    ctx.SystemComponent,
    "realizing_logical_components",
    c.ReferenceSearchingAccessor(
        LogicalComponent, "realized_components", aslist=c.ElementList
    ),
)
c.set_accessor(
    ctx.SystemFunction,
    "realizing_logical_functions",
    c.ReferenceSearchingAccessor(
        LogicalFunction, "realized_system_functions", aslist=c.ElementList
    ),
)
c.set_accessor(
    LogicalFunction,
    "owner",
    c.ReferenceSearchingAccessor(LogicalComponent, "allocated_functions"),
)
c.set_accessor(
    LogicalFunction,
    "packages",
    c.DirectProxyAccessor(
        LogicalFunctionPkg,
        aslist=c.ElementList,
    ),
)
c.set_self_references(
    (LogicalComponent, "components"),
    (LogicalComponentPkg, "packages"),
    (LogicalFunction, "functions"),
    (LogicalFunctionPkg, "packages"),
)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Classes handling Mode/State-Machines and related content."""
from .. import common as c
from . import capellacore, modellingcore

XT_TRAFO = "org.polarsys.capella.core.data.capellacommon:TransfoLink"
XT_ABSTRACT_STATE_REAL = (
    "org.polarsys.capella.core.data.capellacommon:AbstractStateRealization"
)


@c.xtype_handler(None)
class Region(c.GenericElement):
    """A region inside a state machine or state/mode."""

    _xmltag = "ownedRegions"

    states: c.Accessor
    modes: c.Accessor
    transitions: c.Accessor


class AbstractStateMode(c.GenericElement):
    """Common code for states and modes."""

    _xmltag = "ownedStates"

    regions = c.DirectProxyAccessor(Region, aslist=c.ElementList)

    functions: c.Accessor


@c.xtype_handler(None)
class State(AbstractStateMode):
    """A state."""

    entries = c.AttrProxyAccessor(
        c.GenericElement, "entry", aslist=c.MixedElementList
    )
    do_activity = c.AttrProxyAccessor(
        c.GenericElement, "doActivity", aslist=c.MixedElementList
    )
    exits = c.AttrProxyAccessor(
        c.GenericElement, "exit", aslist=c.MixedElementList
    )


@c.xtype_handler(None)
class Mode(AbstractStateMode):
    """A mode."""


@c.xtype_handler(None)
class DeepHistoryPseudoState(AbstractStateMode):
    """A deep history pseudo state."""


@c.xtype_handler(None)
class FinalState(AbstractStateMode):
    """A final state."""


@c.xtype_handler(None)
class ForkPseudoState(AbstractStateMode):
    """A fork pseudo state."""


@c.xtype_handler(None)
class InitialPseudoState(AbstractStateMode):
    """An initial pseudo state."""


@c.xtype_handler(None)
class JoinPseudoState(AbstractStateMode):
    """A join pseudo state."""


@c.xtype_handler(None)
class ShallowHistoryPseudoState(AbstractStateMode):
    """A shallow history pseudo state."""


@c.xtype_handler(None)
class TerminatePseudoState(AbstractStateMode):
    """A terminate pseudo state."""


@c.xtype_handler(None)
class StateMachine(c.GenericElement):
    """A state machine."""

    _xmltag = "ownedStateMachines"

    regions = c.DirectProxyAccessor(Region, aslist=c.ElementList)


@c.xtype_handler(None)
class StateTransition(c.GenericElement):
    r"""A transition between :class:`State`\ s or :class:`Mode`\ s."""

    _xmltag = "ownedTransitions"

    source = c.AttrProxyAccessor(c.GenericElement, "source")
    destination = c.AttrProxyAccessor(c.GenericElement, "target")
    triggers = c.AttrProxyAccessor(
        c.GenericElement, "triggers", aslist=c.MixedElementList
    )
    effects = c.AttrProxyAccessor(
        c.GenericElement, "effect", aslist=c.MixedElementList
    )
    guard = c.AttrProxyAccessor(capellacore.Constraint, "guard")


@c.xtype_handler(None)
class GenericTrace(modellingcore.TraceableElement):
    """A trace between two elements."""

    @property
    def name(self) -> str:  # type: ignore[override]
        """Return the name."""
        direction = ""
        if self.target is not None:
            direction = f" to {self.target.name} ({self.target.uuid})"

        return f"[{type(self).__name__}]{direction}"


c.set_accessor(
    AbstractStateMode,
    "realized_states",
    # FIXME fill in tag
    c.LinkAccessor(
        None,
        XT_ABSTRACT_STATE_REAL,
        aslist=c.ElementList,
        attr="targetElement",
    ),
)
for cls in [
    State,
    Mode,
    DeepHistoryPseudoState,
    FinalState,
    ForkPseudoState,
    InitialPseudoState,
    JoinPseudoState,
    ShallowHistoryPseudoState,
    TerminatePseudoState,
]:
    c.set_accessor(
        cls,
        "realizing_states",
        c.ReferenceSearchingAccessor(
            cls, "realized_states", aslist=c.ElementList
        ),
    )

c.set_accessor(
    Region,
    "states",
    c.RoleTagAccessor(AbstractStateMode._xmltag, aslist=c.ElementList),
)
c.set_accessor(
    Region, "modes", c.DirectProxyAccessor(Mode, aslist=c.ElementList)
)
c.set_accessor(
    Region,
    "transitions",
    c.DirectProxyAccessor(StateTransition, aslist=c.ElementList),
)
c.set_accessor(
    c.GenericElement,
    "traces",
    c.DirectProxyAccessor(GenericTrace, aslist=c.ElementList),
)
